/*
 csv2wav
 
 Converts CSV text data (amplitude and time) into a WAV file. 
 The data are automatically rescaled according to the specifications 
 of the WAV format and can optionnaly be resampled.
  
 Centre for Digital Music, Queen Mary University of London.
 This file copyright 2010-2011 Mathieu Barthet and QMUL.
 
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License as
 published by the Free Software Foundation; either version 2 of the
 License, or (at your option) any later version.  See the file
 COPYING included with this distribution for more information.
 
 Version: 1
*/

#include <cmath>
#include <cstdlib>
#include <cstring>
#include <fstream>
#include <math.h>
#include <iostream>
#include <sstream>
#include <iomanip>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <vector>

#include <samplerate.h> //libsndrate library for resampling
#include <sndfile.h>

using std::cout;
using std::cin;
using std::cerr;
using std::getline;
using std::endl;
using std::string;
using std::vector;
using std::ofstream;
using std::ios;

using namespace std;

#define	VERBOSE

typedef unsigned long DWORD;
typedef unsigned char BYTE;

void print_usage(){
	
	printf("\nUsage :\n");
	printf("\n <time file> <data file> <output file> <(optional) output sample rate> \n\n");
	printf("     <time file> : path of the csv file with the times (e.g. ./time.csv)\n");
	printf("     <data file> : path of the csv file with the data (e.g. ./data.csv)\n");
	printf("     <output file> : path of the output wave file (e.g. ./output.wav)\n");
	printf("     <(optionnal) output sample rate> : the output signal will be resampled according to the specified sample rate\n\n");
	
	printf("Example:\n");
	printf("\n ./csv2wav ./examples/time.csv ./examples/data.csv ./examples/output.wav\n\n");
	
}


/****************************************************************************************/
int main (int argc, char* argv[]) {
	
	char timefile_path[256];
	char datafile_path[256];
	char outputfile_path[256];
	
	float ts;
	float fs;
	int fsout;
	
	//char output_name[256];
	
	char timeline[256];
	char dataline[256];
		
	float *time;
	float *data;
	
	//Output Sound file
	SNDFILE* outputfile;
	struct SF_INFO* osf_info;
	sf_count_t wf;
	
	int i;
	int nl_time, nl_data; //number of lines of the time and data files, respectively
	
	int Nframes; //number of frames to be written in the wave file
			
	char temp[256]; //temporary array to count the number of lines in the files
	
	//struct stat st;
	
	float ampcoeff = 0.999; //amplitude coefficient to rescale the signal before saving the wave file
	
	bool resampling = false;
	
	//resamping args
	SRC_DATA *resampledata; //structure with the data for resampling
		
	float  *data_out ;	
	long   input_frames, output_frames ;
	double src_ratio ;
	
	int converter_type = SRC_SINC_BEST_QUALITY; //see http://www.mega-nerd.com/SRC/api_misc.html#Converters
	int channels = 1;
	
	/*************************
	 Argument processing
	 **************************/
	
	if ((argc != 4)&&(argc != 5)) {
		cout << argc << endl;
		print_usage();
		exit(EXIT_FAILURE);
	}
	
	strcpy(timefile_path,argv[1]);
	strcpy(datafile_path,argv[2]);
	strcpy(outputfile_path,argv[3]);
	
	//new sampling rate (optionnal)
	if (argc == 5){
		resampling = true;
		fsout = atoi(argv[4]);
	}
			
	//Reads the data from the csv files
	
	//time signal
	ifstream timefile(timefile_path,ios::in|ios::binary);
	if (!timefile) {
		cout << "Error opening time file" << endl;
		exit(-1);
	}
	
	//counts the number of lines
	nl_time = 0;
	//float f;
	timefile.getline(temp,256,'\n');
	while (timefile.gcount()!=0){
		timefile.getline(temp,256,'\n');
		nl_time ++;
	}
	cout << "Number of lines (time file):  "<< nl_time << endl;
	
	//rewinds the pointer to the file
	timefile.clear();
	timefile.seekg(0,ios_base::beg);
	
	//allocates memory
	time = (float *)malloc(sizeof(float)*nl_time);	
	if (time==NULL){
		cerr << "ERROR: Unable to allocate memory (time)."
		<< endl;
		return 0;
	}
	
	for (i=0;i<nl_time;i++)
	{	//Reads the first line
		timefile.getline(timeline,256,'\n');
		//retrieves the time
		sscanf(timeline,"%f",time+i);
		//cout << time[i] << endl;
	}		
	
	cout << "Last measurement time: " << time[nl_time-1] << " s" << endl; 
	
	//data signal
	ifstream datafile(datafile_path,ios::in|ios::binary);
	if (!datafile) {
		cout << "Error opening datafile" << endl;
		exit(-1);
	}
	
	//counts the number of lines
	nl_data = 0;
	datafile.getline(temp,256,'\n');
	while (datafile.gcount()!=0){
		datafile.getline(temp,256,'\n');
		nl_data++;
	}
	cout << "Number of lines (data file): " << nl_data << endl;
	
	if (nl_data!=nl_time){
		cerr << "ERROR: the number of lines in the time and data files is not the same."
		<< endl;
		return 0;
	}
	
	//rewinds the pointer to the file
	datafile.clear();
	datafile.seekg(0,ios_base::beg);
	
	//allocates memory
	data = (float *)malloc(sizeof(float)*nl_data);	
	if (data==NULL){
		cerr << "ERROR: Unable to allocate memory (data)."
		<< endl;
		return 0;
	}
	
	for (i=0;i<nl_data;i++)
	{	//Reads the first line
		datafile.getline(dataline,256,'\n');
		//retrieves the data
		sscanf(dataline,"%f",data+i);
		//cout << data[i] << endl;
	}		
	
	// find the actual sample rate
	ts = time[1]-time[0]; //assumes that the sampling rate in the input signal is constant
	//fs = round(1/ts);
	fs = 1/ts;
	
	Nframes = nl_data; //number of frames to write in the wave file
	
	if (resampling){ //resampling
		
		cout << "Starts resampling" << endl;
		
		resampledata = (SRC_DATA *)malloc(sizeof(SRC_DATA));
		if (resampledata==NULL){
			cerr << "ERROR: Unable to allocate memory (resampledata)."
			<< endl;
			return 0;
		}
		
		cout << "fs: " << fs << " - fsout: " << fsout << endl; 
		
		src_ratio = (double) fsout/fs;
		
		input_frames = (long) nl_data;
				
		float dur_in = time[nl_time-1];
				
		output_frames = dur_in*fsout;
		
		cout << "Resampling ratio: " << src_ratio << endl;
					
		data_out = (float *)malloc(sizeof(float)*output_frames);
		if (data_out==NULL){
			cerr << "ERROR: Unable to allocate memory (data_out)."
			<< endl;
			return 0;
		}		
		
		cout << "Number of output frames: " << output_frames << endl;
		
		cout << "Type of converter: " << converter_type << endl;
		
		resampledata->data_in = data; //input data
		resampledata->input_frames = input_frames;
		resampledata->data_out = data_out;
		resampledata->output_frames = output_frames;
		resampledata->src_ratio = src_ratio;
				
		int resample_err = src_simple(resampledata, converter_type, channels) ;
		
		if (resample_err!=0) {			
			cout << "ERROR: during resampling (src_simple): " << src_strerror(resample_err) << endl;
			cout << resampledata->input_frames_used << endl;
			return 0;
		}
		
		cout << "Number of input frames used: " << resampledata->input_frames_used << endl;
		cout << "Number of output frames generated: " << resampledata->output_frames_gen << endl;
		
		data = resampledata->data_out;
		
		free(resampledata);
		
		//copy the new sample frequency and number of frames
		fs = fsout;
		
		Nframes = resampledata->output_frames_gen;
	}
	
	//removes mean
	float datasum = 0;
	for (i=0;i<Nframes;i++)
	{	
		datasum = datasum + data[i];
	}
	
	float datamean = datasum/Nframes;
	
	for (i=0;i<Nframes;i++)
	{	
		data[i] = data[i] - datamean;
	}
	
	//rescale
	//finds the maximum
	float datamax = 0;
	for (i=0;i<Nframes;i++)
	{	if (abs(data[i]) > datamax)
	{	
		datamax = abs(data[i]);
	}
	}
	//rescale
	for (i=0;i<Nframes;i++)
	{
		data[i] = data[i]*ampcoeff/datamax;
		
	}
		
	//cout << "The output sampling rate is: " << fs << " Hz." << endl;
	
	//Writes the output mono wave file 
	
	osf_info = (struct SF_INFO *)malloc(sizeof(struct SF_INFO));
	if (osf_info==NULL){
		cerr << "ERROR: Unable to allocate memory (osf_info)."
		<< endl;
		return 0;
	}
	
	sf_count_t oframes = (sf_count_t) Nframes;
	
	osf_info->format = SF_FORMAT_WAV | SF_FORMAT_PCM_16;
	osf_info->channels = 1;
	osf_info->frames = oframes;
	osf_info->samplerate = (int) round(fs);
	osf_info->sections = 1;
	osf_info->seekable = true;
	
	printf("Writing file %s at sampling rate %i Hz.\n", outputfile_path, (int) round(fs));
	
	//outputfile = sf_open(outputfile_path,SFM_RDWR,osf_info);
	outputfile = sf_open(outputfile_path,SFM_WRITE,osf_info);
	
	if (outputfile!=NULL){
		
		wf = sf_writef_float(outputfile, data, Nframes);
		
		if (wf!=Nframes){
			printf("Problem writing file (length = %i, wf = %i).\n",Nframes,(int) wf);
			exit(-1);
		}
	}else{
		printf("Problem sf_open: no output file specified.\n");
		exit(-1);
	}
	
	sf_close(outputfile);
	
	
	timefile.close();
	datafile.close();
		
	free(time);
	free(data);
	free(osf_info);
}
